/* 
 * File:   i2c.h
 * Author: tim
 *
 * Created on October 31, 2017, 5:59 PM
 */

#define USE_AND_OR	// To enable AND_OR mask setting for I2C. 
#include "plib.h"
#include "tft_master.h"
#define IMU_ADDRESS 0xd0
#define LUX_ADDRESS 0x72
#define ACCEL_XOUT_H 0x3b
#define ACCEL_XOUT_L 0x3c
#define ACCEL_YOUT_H 0x3d
#define ACCEL_YOUT_L 0x3e
#define ACCEL_ZOUT_H 0x3f
#define ACCEL_ZOUT_L 0x40
#define GYRO_XOUT_H 0x43
#define GYRO_XOUT_L 0x44
#define GYRO_YOUT_H 0x45
#define GYRO_YOUT_L 0x46
#define GYRO_ZOUT_H 0x47
#define GYRO_ZOUT_L 0x48

// IMU Register Definitions
#define MPU6050_RA_PWR_MGMT_1       0x6b
#define MPU6050_RA_GYRO_CONFIG      0x1b
#define MPU6050_RA_ACCEL_CONFIG     0x1c

#define GYRO_SCALE 131.0 // lsb/(degrees/second)

#define _tsl2561AutoGain                0
#define TSL2561_INTEGRATIONTIME_13MS    0x01 // should be 0
#define TSL2561_GAIN_1X                 0x01
#define TSL2561_COMMAND_BIT             0x80 // must be 1
#define TSL2561_REGISTER_TIMING         0x01
#define TSL2561_WORD_BIT                0x20
#define TSL2561_REGISTER_CHAN0_LOW      0x0C
#define TSL2561_REGISTER_CHAN1_LOW      0x0E
#define TSL2561_REGISTER_CHAN0_HIGH     0x0D
#define TSL2561_REGISTER_CHAN1_HIGH     0x0F
#define TSL2561_REGISTER_CONTROL        0x00
#define TSL2561_CH_SCALE                0x7517
#define TSL2561_LUX_CHSCALE             10
#define TSL2561_LUX_RATIOSCALE          9
#define TSL2561_LUX_LUXSCALE            14
#define TSL2561_LUX_K1T                 (0x0040)  // 0.125 * 2^RATIO_SCALE
#define TSL2561_LUX_B1T                 (0x01f2)  // 0.0304 * 2^LUX_SCALE
#define TSL2561_LUX_M1T                 (0x01be)  // 0.0272 * 2^LUX_SCALE
#define TSL2561_LUX_K2T                 (0x0080)  // 0.250 * 2^RATIO_SCALE
#define TSL2561_LUX_B2T                 (0x0214)  // 0.0325 * 2^LUX_SCALE
#define TSL2561_LUX_M2T                 (0x02d1)  // 0.0440 * 2^LUX_SCALE
#define TSL2561_LUX_K3T                 (0x00c0)  // 0.375 * 2^RATIO_SCALE
#define TSL2561_LUX_B3T                 (0x023f)  // 0.0351 * 2^LUX_SCALE
#define TSL2561_LUX_M3T                 (0x037b)  // 0.0544 * 2^LUX_SCALE
#define TSL2561_LUX_K4T                 (0x0100)  // 0.50 * 2^RATIO_SCALE
#define TSL2561_LUX_B4T                 (0x0270)  // 0.0381 * 2^LUX_SCALE
#define TSL2561_LUX_M4T                 (0x03fe)  // 0.0624 * 2^LUX_SCALE
#define TSL2561_LUX_K5T                 (0x0138)  // 0.61 * 2^RATIO_SCALE
#define TSL2561_LUX_B5T                 (0x016f)  // 0.0224 * 2^LUX_SCALE
#define TSL2561_LUX_M5T                 (0x01fc)  // 0.0310 * 2^LUX_SCALE
#define TSL2561_LUX_K6T                 (0x019a)  // 0.80 * 2^RATIO_SCALE
#define TSL2561_LUX_B6T                 (0x00d2)  // 0.0128 * 2^LUX_SCALE
#define TSL2561_LUX_M6T                 (0x00fb)  // 0.0153 * 2^LUX_SCALE
#define TSL2561_LUX_K7T                 (0x029a)  // 1.3 * 2^RATIO_SCALE
#define TSL2561_LUX_B7T                 (0x0018)  // 0.00146 * 2^LUX_SCALE
#define TSL2561_LUX_M7T                 (0x0012)  // 0.00112 * 2^LUX_SCALE
#define TSL2561_LUX_K8T                 (0x029a)  // 1.3 * 2^RATIO_SCALE
#define TSL2561_LUX_B8T                 (0x0000)  // 0.000 * 2^LUX_SCALE
#define TSL2561_LUX_M8T                 (0x0000)  // 0.000 * 2^LUX_SCALE

#define _I2C1CON_ACKDT_POSITION                  0x00000005
#define _I2C1CON_ACKDT_MASK                      0x00000020
#define _I2C1CON_ACKDT_LENGTH                    0x00000001

#define _I2C1CON_ACKEN_POSITION                  0x00000004
#define _I2C1CON_ACKEN_MASK                      0x00000010
#define _I2C1CON_ACKEN_LENGTH                    0x00000001

#define ack() (I2C1CONCLR = _I2C1CON_ACKDT_MASK, I2C1CONSET = _I2C1CON_ACKEN_MASK)

float X_ACC_OFF, Y_ACC_OFF, Z_ACC_OFF, X_GYRO_OFF, Y_GYRO_OFF, Z_GYRO_OFF;

// Wait by executing nops
void i2c_wait(unsigned int cnt) 
{
  while (--cnt) 
  {
    asm( "nop");
    asm( "nop");
  }
}

void init_gyro() 
{
  i2c_write_bits(IMU_ADDRESS, MPU6050_RA_PWR_MGMT_1, 2, 1, 3);
  i2c_write_bits(IMU_ADDRESS, MPU6050_RA_GYRO_CONFIG, 4, 0, 2);
  i2c_write_bits(IMU_ADDRESS, MPU6050_RA_ACCEL_CONFIG, 4, 0, 2);
  i2c_write_bits(IMU_ADDRESS, MPU6050_RA_PWR_MGMT_1, 6, 0, 1);
}

// Write a number of chars from data specified by num to the specified address
void i2c_write(char d_address, char address, char *data, int num) 
{
  int i;
  char i2c_header[2];
  
  // device address & WR
  i2c_header[0] = d_address | 0; 
  
  // register address
  i2c_header[1] = address; 
  
  // Send the Start Bit
  StartI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 
  
  for (i = 0; i < num + 2; i++) 
  {
    if (i < 2)
      MasterWriteI2C1(i2c_header[i]);
    else
      MasterWriteI2C1(data[i - 2]);
      // Wait to complete
      IdleI2C1(); 

    // ACKSTAT is 0 when slave acknowledge, 
    // if 1 then slave has not acknowledge the data.
    if (I2C1STATbits.ACKSTAT)
      break;
  }
  // Send the Stop condition
  StopI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 
}

// Read a char from the register specified by address
char i2c_read(char d_address, char address) 
{
  char i2c_header[2];
  
  // device address & WR
  i2c_header[0] = (d_address | 0); 
  
  // register address
  i2c_header[1] = address; 

  // Send the Start Bit
  StartI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 

  int i;
  for (i = 0; i < 2; i++) 
  {
    MasterWriteI2C1(i2c_header[i]);
    
    // Wait to complete
    IdleI2C1(); 

    // ACKSTAT is 0 when slave acknowledge, 
    // if 1 then slave has not acknowledge the data.
    if (I2C1STATbits.ACKSTAT) 
    {
      break;
    }
  }
  // Send the Restart condition
  RestartI2C1(); 
  i2c_wait(10);
  
  // wait for this bit to go back to zero
  IdleI2C1();

  // transmit read command
  MasterWriteI2C1(d_address | 1); 
  
  // Wait to complete
  IdleI2C1(); 

  // read some bytes back
  char data = MasterReadI2C1();

  // Wait to complete
  IdleI2C1(); 

  // Send the Stop condition
  StopI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 
  return data;
}

// Write a number of chars from data specified by num to the specified address

void i2c_write_bits(char d_address, char address, int bit_start, char data, int length) 
{
  char b = i2c_read(d_address, address);
  if (b != 0) 
  {
    char mask = ((1 << length) - 1) << (bit_start - length + 1);
    data <<= (bit_start - length + 1);
    data &= mask;
    b &= ~(mask);
    b |= data;
    char data_r[] = {b};
    i2c_write(d_address, address, data_r, 1);
  }
}

// Read three-axis accelerometer and three-axis gyroscope from MPU 6050
// Return values in array pointed to by values

void read_imu_values(float* values) 
{
  int xAccelH = (int) i2c_read(IMU_ADDRESS, ACCEL_XOUT_H);
  int xAccelL = (int) i2c_read(IMU_ADDRESS, ACCEL_XOUT_L);
  int yAccelH = (int) i2c_read(IMU_ADDRESS, ACCEL_YOUT_H);
  int yAccelL = (int) i2c_read(IMU_ADDRESS, ACCEL_YOUT_L);
  int zAccelH = (int) i2c_read(IMU_ADDRESS, ACCEL_ZOUT_H);
  int zAccelL = (int) i2c_read(IMU_ADDRESS, ACCEL_ZOUT_L);
//  values[0] = (float) ((xAccelH << 8) + xAccelL) - X_ACC_OFF;
//  values[1] = (float) ((yAccelH << 8) + yAccelL) - Y_ACC_OFF;
//  values[2] = (float) ((zAccelH << 8) + zAccelL) - Z_ACC_OFF;
  values[0] = (float) ((xAccelH << 8) + xAccelL)/16384 - X_ACC_OFF;
  values[1] = (float) ((yAccelH << 8) + yAccelL)/16384 - Y_ACC_OFF;
  values[2] = (float) ((zAccelH << 8) + zAccelL)/16384 - Z_ACC_OFF;
}

// Sample the gyroscope over an extended period and average
// in order to find and subtract the bias
void calibrate() 
{
  int numSamples = 500;
  int i = 0;
  float xAccSum, yAccSum, zAccSum;
  float values[6];

  for (i = 0; i < numSamples; i++) 
  {
    read_imu_values(values);
    
    // Parse the IMU data
    float xAcc = values[0];
    float yAcc = values[1];
    float zAcc = values[2];
    
    xAccSum += xAcc;
    yAccSum += yAcc;
    zAccSum += zAcc;
  }
  X_ACC_OFF += xAccSum / numSamples;
  X_ACC_OFF = X_ACC_OFF + 1;
  Y_ACC_OFF += yAccSum / numSamples;
  Z_ACC_OFF += zAccSum / numSamples;
}

void init_luminosity() 
{
  char data[] = {0x00};
  i2c_write(LUX_ADDRESS, TSL2561_COMMAND_BIT | TSL2561_REGISTER_TIMING, data, 1);

  data[0] = 0x03;
  i2c_write(LUX_ADDRESS, TSL2561_COMMAND_BIT | TSL2561_REGISTER_CONTROL, data, 1);
}

unsigned int read16(char command) 
{
  unsigned int toRet;
  char i2c_header[2];
  int i;
  
  // device address & WR
  i2c_header[0] = (LUX_ADDRESS | 0); 
  
  // register address
  i2c_header[1] = command; 

  // Send the Start Bit
  StartI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 

  for (i = 0; i < 2; i++) 
  {
    MasterWriteI2C1(i2c_header[i]);
    
    // Wait to complete
    IdleI2C1(); 

    // ACKSTAT is 0 when slave acknowledge, 
    // if 1 then slave has not acknowledge the data.
    while (I2C1STATbits.ACKSTAT) 
    {
    }  
  }

  // Send the Restart condition
  StartI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 

  // transmit read command
  MasterWriteI2C1(0x73); 
  
  // Wait to complete
  IdleI2C1(); 

  while (I2C1STATbits.ACKSTAT) 
  {
  }

  unsigned int low = MasterReadI2C1();
  
  // Wait to complete
  IdleI2C1(); 
  AckI2C1();

  IdleI2C1();
  unsigned int high = MasterReadI2C1();
  NotAckI2C1();
  IdleI2C1();

  // Send the Stop condition
  StopI2C1(); 
  IdleI2C1();

  toRet = high << 8;
  toRet |= low;
  return toRet;
}

void get_luminosity(short *values) 
{
  unsigned int ch0= (read16(0xAC) * TSL2561_CH_SCALE) >> TSL2561_LUX_CHSCALE;
  unsigned int ch1= (read16(0xAE) * TSL2561_CH_SCALE) >> TSL2561_LUX_CHSCALE;
    
  unsigned long ratio1 = 0;
  if (ch0 != 0) ratio1 = (ch1 << (TSL2561_LUX_RATIOSCALE+1)) / ch0;
  unsigned long ratio = (ratio1 + 1) >> 1;
  unsigned int b, m;
   
  if ((ratio >= 0) && (ratio <= TSL2561_LUX_K1T))
  {
    b=TSL2561_LUX_B1T; 
    m=TSL2561_LUX_M1T;
  }
  else if (ratio <= TSL2561_LUX_K2T)
  {
    b=TSL2561_LUX_B2T; 
    m=TSL2561_LUX_M2T;
  }
  else if (ratio <= TSL2561_LUX_K3T)
  {
    b=TSL2561_LUX_B3T; 
    m=TSL2561_LUX_M3T;
  }
  else if (ratio <= TSL2561_LUX_K4T)
  {
    b=TSL2561_LUX_B4T; 
    m=TSL2561_LUX_M4T;
  }
  else if (ratio <= TSL2561_LUX_K5T)
  {
    b=TSL2561_LUX_B5T; 
    m=TSL2561_LUX_M5T;
  }
  else if (ratio <= TSL2561_LUX_K6T)
  {
    b=TSL2561_LUX_B6T; 
    m=TSL2561_LUX_M6T;
  }
  else if (ratio <= TSL2561_LUX_K7T)
  {
    b=TSL2561_LUX_B7T; 
    m=TSL2561_LUX_M7T;
  }
  else if (ratio > TSL2561_LUX_K8T)
  {
    b=TSL2561_LUX_B8T; 
    m=TSL2561_LUX_M8T;
  }
  unsigned long temp;
  temp = ((ch0 * b) - (ch1 * m));
  temp += (1 << (TSL2561_LUX_LUXSCALE-1));
  unsigned int lux = temp >> TSL2561_LUX_LUXSCALE;
  values[0] = lux;
}

unsigned int temp_get(char command) 
{
  unsigned int toRet;
  char i2c_header[2];

  // device address & WR
  i2c_header[0] = (0x80 | 0); 
    
  // register address
  i2c_header[1] = command; 

  // Send the Start Bit
  StartI2C1(); 
    
  // Wait to complete
  IdleI2C1(); 

  int i;
  for (i = 0; i < 2; i++) 
  {
    MasterWriteI2C1(i2c_header[i]);
    
    // Wait to complete
    IdleI2C1(); 

    // ACKSTAT is 0 when slave acknowledge, 
    // if 1 then slave has not acknowledge the data.
    while (I2C1STATbits.ACKSTAT) 
    {
    }
  }

  // Send the Restart condition
  StartI2C1(); 
  
  // Wait to complete
  IdleI2C1(); 

  // transmit read command
  MasterWriteI2C1(0x80 | 1); 
  
  // Wait to complete
  IdleI2C1(); 

  while (I2C1STATbits.ACKSTAT) 
  {
  }

  unsigned int high = MasterReadI2C1();

  // Wait to complete
  IdleI2C1(); 
  AckI2C1();

  IdleI2C1();
  unsigned int low = MasterReadI2C1();
  NotAckI2C1();
  IdleI2C1();

  // Send the Stop condition
  StopI2C1(); 
  IdleI2C1();

  low &= 0xFC;
  toRet = high << 8 | low;
  return toRet;
}

void measure_temp_humidity(float *values) 
{
  unsigned int temp, humidity;
  temp = temp_get(0xE3);
  humidity = temp_get(0xE5);
  float result = ((175.72 * temp) / 65536) - 46.85;
  result = (result * 1.8) + 32.0;
  values[0] = result;
  result = ((125.0 * humidity) / 65536) - 6;
  values[1] = result;
}
